@@ -19,8 +19,6 @@ from typing import List, Optional, Tuple
 import numpy as np
 from scipy.linalg import fractional_matrix_power
 
-from braket.default_simulator.linalg_utils import controlled_unitary
-
 
 class Operation(ABC):
     """
@@ -50,6 +48,7 @@ class GateOperation(Operation, ABC):
         self._ctrl_modifiers = ctrl_modifiers
         self._power = power
 
+    @property
     @abstractmethod
     def _base_matrix(self) -> np.ndarray:
         """np.ndarray: The matrix representation of the operation."""
@@ -61,13 +60,14 @@ class GateOperation(Operation, ABC):
             unitary = np.linalg.matrix_power(unitary, int(self._power))
         else:
             unitary = fractional_matrix_power(unitary, self._power)
-
-        for mod in self._ctrl_modifiers:
-            unitary = controlled_unitary(unitary, negctrl=mod)
         return unitary
 
     def __eq__(self, other):
-        return self.targets == other.targets and np.allclose(self.matrix, other.matrix)
+        possible_parameters = "_angle", "_angle_1", "_angle_2"
+        return self.targets == other.targets and all(
+            getattr(self, param, None) == getattr(other, param, None)
+            for param in possible_parameters
+        )
 
 
 class KrausOperation(Operation, ABC):
