@@ -36,6 +36,10 @@ def apply_operations(
     """
     for operation in operations:
         matrix = operation.matrix
-        targets = operation.targets
-        state = multiply_matrix(state, matrix, targets)
+        all_targets = operation.targets
+        num_ctrl = len(operation._ctrl_modifiers)
+        control_state = operation._ctrl_modifiers
+        controls = all_targets[:num_ctrl]
+        targets = all_targets[num_ctrl:]
+        state = multiply_matrix(state, matrix, targets, controls, control_state)
     return state
